# -*- coding: utf-8 -*-

from warnings import warn
from optparse import OptionGroup

from . import forms
from .pageobject import Page
from .router import add_route
from .utils import re_raise_exc
from .extension import assertion
from .pageobject import PageObject
from .extension import inject_driver
from .pageobject import PageObjectProxy
from .extension import case_of_browsers
from .extension import SeleniumAssertion
from .query import Contains as _Contains
from .extension import SeleniumCase as Case
from .extension import SeleniumSuite as Suite
from .query import QueryObject as _QueryObject


CONFIG_KEY = 'SELENIUM_EX'


def __add_options__(parser):
    group = OptionGroup(parser, 'Selenium extension options')

    group.add_option(
        '--selenium-browser',
        dest='SELENIUM_BROWSERS',
        action='append',
        default=[],
        help='Browser names for run cases.',
    )
    group.add_option(
        '--selenium-project-url',
        dest='SELENIUM_PROJECT_URL',
        default=None,
        help='Base URL of your project.',
    )
    group.add_option(
        '--selenium-remote',
        dest='SELENIUM_REMOTE',
        action='store_true',
        default=False,
        help='Use remote server only.'
    )
    group.add_option(
        '--selenium-no-remote',
        dest='SELENIUM_NO_REMOTE',
        action='store_true',
        default=False,
        help='Use local drivers only.'
    )
    group.add_option(
        '--selenium-polling',
        dest='SELENIUM_POLLING',
        type=float,
        default=None,
        help='Polling timeout.',
    )
    group.add_option(
        '--selenium-polling-delay',
        dest='SELENIUM_POLLING_DELAY',
        type=float,
        default=None,
        help='Polling delay.',
    )
    group.add_option(
        '--selenium-window-size',
        dest='SELENIUM_WINDOW_SIZE',
        default=None,
        help='Window size. For example: 340x480',
    )

    parser.add_option_group(group)


def __install__(program):
    from .extension import EX_NAME
    from .extension import Selenium

    config = program.config.get(CONFIG_KEY, {})

    if program.config.SELENIUM_PROJECT_URL:
        config['PROJECT_URL'] = program.config.SELENIUM_PROJECT_URL

    if program.config.SELENIUM_NO_REMOTE:
        config['USE_REMOTE'] = False

    if program.config.SELENIUM_REMOTE:
        config['USE_REMOTE'] = True

    if program.config.SELENIUM_POLLING:
        config['POLLING_TIMEOUT'] = program.config.SELENIUM_POLLING

    if program.config.SELENIUM_POLLING_DELAY:
        config['POLLING_DELAY'] = program.config.SELENIUM_POLLING_DELAY

    if program.config.SELENIUM_WINDOW_SIZE:
        w, h = program.config.SELENIUM_WINDOW_SIZE.split('x')
        config['WINDOW_SIZE'] = (int(w), int(h))

    if (config['POLLING_TIMEOUT'] and config['IMPLICITLY_WAIT']) \
            and config['POLLING_TIMEOUT'] >= config['IMPLICITLY_WAIT']:
        warn_message = 'POLLING_TIMEOUT >= IMPLICITLY_WAIT it will be work so slowly. ' \
                       'Please set IMPLICITLY_WAIT larger POLLING_TIMEOUT'
        warn(warn_message, RuntimeWarning)

    program.shared_extension(
        EX_NAME, Selenium, args=(config, ),
    )


class _Query(object):
    """
    For usability only.

    query('div', _class=query.contains('name'))
    """

    @property
    def contains(self):
        return _Contains

    def __call__(self, tag, **selector):
        return _QueryObject(tag, **selector)


query = _Query()


__all__ = (
    'Case',
    'Page',
    'forms',
    'query',
    'Suite',
    'add_route',
    'assertion',
    'PageObject',
    're_raise_exc',
    'inject_driver',
    'PageObjectProxy',
    'case_of_browsers',
    'SeleniumAssertion',
)



# -*- coding: utf-8 -*-

import os
import json
import logging
from functools import wraps

try:
    from httplib import HTTPException
except ImportError:  # please python 3
    from http.client import HTTPException

from selenium.common.exceptions import StaleElementReferenceException
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities

from ... import case
from ... import steps
from . import drivers
from ... import suite
from ... import reason
from ... import runnable
from ...utils import pyv
from .proxy import WebDriverProxy
from .utils import random_file_name
from .polling import POLLING_EXCEPTIONS
from ...utils.common import waiting_for
from .exceptions import SeleniumExError


logger = logging.getLogger(__name__)


EX_NAME = 'selenium'

GET_DRIVER_TIMEOUT = 10
DEFAULT_POLLING_TIMEOUT = 10
DEFAULt_POLLING_DELAY = None
DEFAULT_MAXIMIZE_WINDOW = True
DEFAULT_BROWSER = drivers.FIREFOX

DRIVER_TO_CAPABILITIES = {
    drivers.OPERA: DesiredCapabilities.OPERA,
    drivers.CHROME: DesiredCapabilities.CHROME,
    drivers.FIREFOX: DesiredCapabilities.FIREFOX,
    drivers.PHANTOMJS: DesiredCapabilities.PHANTOMJS,
    drivers.IE: DesiredCapabilities.INTERNETEXPLORER,
}


def get_capabilities(driver_name):
    """
    Get capabilities of driver

    :param driver_name: driver name
    :type driver_name: str
    """
    try:
        return DRIVER_TO_CAPABILITIES[driver_name]
    except KeyError:
        raise SeleniumExError(
            'Capabilities for driver "{}" is not found'.find(driver_name),
        )


def create_browser(selenium, driver):
    browser = WebDriverProxy(
        driver,
        config=BrowserConfig(selenium, driver),
    )
    browser.reason_storage['browser name'] = selenium.browser_name
    return browser


def save_logs(case, selenium):
    log_file_name = lambda log_type: '{}-{}({}:{}).log'.format(
        selenium.browser_name,
        log_type,
        runnable.class_name(case),
        runnable.method_name(case),
    )

    def log_to_string(log):
        def create_string_from_item(item, tab=0, break_line=False):
            if isinstance(item, pyv.basestring):
                try:
                    item = json.loads(item)
                except BaseException:
                    pass

            if isinstance(item, dict):
                return ('\n' if break_line else '') + u'\n'.join(
                    u'{}{}: {}'.format(
                        ('  ' * tab), k, create_string_from_item(v, tab=(tab + 1), break_line=True)
                    ) for k, v in item.items()
                )
            elif isinstance(item, (list, tuple)):
                return ('\n' if break_line else '') + u'\n'.join(
                    u'{}{}'.format(
                        ('  ' * tab), create_string_from_item(i, tab=(tab + 1), break_line=True),
                    ) for i in item
                )
            else:
                item = pyv.unicode_string(item)

            return item.strip()

        string = create_string_from_item(log).strip()

        if pyv.IS_PYTHON_2:
            return string.encode('utf-8')
        return string

    with selenium.browser.disable_polling():
        for log_type in selenium.browser.log_types:
            log = selenium.browser.get_log(log_type)
            if not log:
                continue

            log_file_path = os.path.join(
                selenium.browser.config.LOGS_PATH,
                log_file_name(log_type),
            )

            try:
                with open(log_file_path, 'w') as f:
                    f.write(log_to_string(log))
            except BaseException as error:
                logger.error(error, exc_info=True)


class BrowserConfig(object):

    def __init__(self, selenium, browser):
        self.__browser = browser

        self.LOGS_PATH = selenium.config.get('LOGS_PATH')
        self.PROJECT_URL = selenium.config.get('PROJECT_URL')

        self.POLLING_DELAY = selenium.config.get(
            'POLLING_DELAY', DEFAULt_POLLING_DELAY,
        )
        self.POLLING_TIMEOUT = selenium.config.get(
            'POLLING_TIMEOUT', DEFAULT_POLLING_TIMEOUT,
        )

        self.__SCRIPT_TIMEOUT = None
        self.SCRIPT_TIMEOUT = selenium.config.get('SCRIPT_TIMEOUT')

        self.__IMPLICITLY_WAIT = None
        self.IMPLICITLY_WAIT = selenium.config.get('IMPLICITLY_WAIT')

        self.__WINDOW_SIZE = None
        self.WINDOW_SIZE = selenium.config.get('WINDOW_SIZE')

        self.__MAXIMIZE_WINDOW = False
        self.MAXIMIZE_WINDOW = selenium.config.get(
            'MAXIMIZE_WINDOW', DEFAULT_MAXIMIZE_WINDOW,
        )

    @property
    def SCRIPT_TIMEOUT(self):
        return self.__SCRIPT_TIMEOUT

    @SCRIPT_TIMEOUT.setter
    def SCRIPT_TIMEOUT(self, value):
        if value is not None:
            self.__SCRIPT_TIMEOUT = value
            self.__browser.set_script_timeout(value)

    @property
    def IMPLICITLY_WAIT(self):
        return self.__IMPLICITLY_WAIT

    @IMPLICITLY_WAIT.setter
    def IMPLICITLY_WAIT(self, value):
        if value is not None:
            self.__IMPLICITLY_WAIT = value
            self.__browser.implicitly_wait(value)

    @property
    def WINDOW_SIZE(self):
        return self.__WINDOW_SIZE

    @WINDOW_SIZE.setter
    def WINDOW_SIZE(self, value):
        if value:
            self.__WINDOW_SIZE = value
            self.__browser.set_window_size(
                *self.WINDOW_SIZE
            )

    @property
    def MAXIMIZE_WINDOW(self):
        return self.__MAXIMIZE_WINDOW

    @MAXIMIZE_WINDOW.setter
    def MAXIMIZE_WINDOW(self, value):
        if value and not self.__WINDOW_SIZE:
            self.__MAXIMIZE_WINDOW = value
            self.__browser.maximize_window()


class Selenium(object):

    def __init__(self, config):
        self.__config = config

        self.__browser = None
        self.__browser_name = None

    def __enter__(self):
        self.start()
        return self.__browser

    def __exit__(self, *args, **kwargs):
        self.stop()

    @property
    def config(self):
        return self.__config

    @property
    def browser(self):
        return self.__browser

    @property
    def browser_name(self):
        return self.__browser_name

    def start(self, *args, **kwargs):
        if self.__browser:
            return

        self.get_browser(*args, **kwargs)

    def stop(self):
        if self.__browser is None:
            return

        with self.__browser.disable_polling():
            self.__browser.quit()

        self.__browser = None

    def remote(self, driver_name):
        driver_name = driver_name.lower()
        remote_config = self.__config.get('REMOTE')

        if not remote_config:
            raise SeleniumExError('settings of remote not found in config')

        logger.debug(
            'Remote config: {}'.format(str(remote_config)),
        )

        options = remote_config.get('OPTIONS', {})
        capabilities = get_capabilities(driver_name)
        capabilities.update(
            remote_config['CAPABILITIES'][driver_name],
        )

        driver = drivers.RemoteWebDriver(
            desired_capabilities=capabilities,
            **options
        )
        self.__browser = create_browser(self, driver)

        return self.__browser

    def ie(self):
        ie_config = self.__config.get('IE')

        if not ie_config:
            raise SeleniumExError('settings of ie browser not found in config')

        logger.debug(
            'Ie config: {}'.format(str(ie_config)),
        )

        driver = drivers.IeWebDriver(**ie_config)
        self.__browser = create_browser(self, driver)

        return self.__browser

    def chrome(self):
        """
        :rtype: selenium.webdriver.chrome.webdriver.WebDriver
        """
        chrome_config = self.__config.get('CHROME')

        if not chrome_config:
            raise SeleniumExError('settings of chrome browser not found in config')

        logger.debug(
            'Chrome config: {}'.format(str(chrome_config)),
        )

        driver = drivers.ChromeWebDriver(**chrome_config)
        self.__browser = create_browser(self, driver)

        return self.__browser

    def firefox(self):
        firefox_config = self.__config.get('FIREFOX', {})

        logger.debug(
            'Firefox config: {}'.format(str(firefox_config)),
        )

        driver = drivers.FirefoxWebDriver(**firefox_config)
        self.__browser = create_browser(self, driver)

        return self.__browser

    def phantomjs(self):
        phantom_config = self.__config.get('PHANTOMJS')

        if not phantom_config:
            raise SeleniumExError('settings of phantom js not found in config')

        logger.debug(
            'PhantomJS config: {}'.format(str(phantom_config)),
        )

        driver = drivers.PhantomJSWebDriver(**phantom_config)
        self.__browser = create_browser(self, driver)

        return self.__browser

    def opera(self):
        opera_config = self.__config.get('OPERA')

        if not opera_config:
            raise SeleniumExError('settings of opera browser not found in config')

        logger.debug(
            'Opera config: {}'.format(str(opera_config)),
        )

        driver = drivers.OperaWebDriver(**opera_config)
        self.__browser = create_browser(self, driver)

        return self.__browser

    def get_browser(self, browser_name=None, timeout=None, delay=None):
        if self.__browser:
            return self.__browser

        self.__browser_name = (
            browser_name or self.__config.get('DEFAULT_BROWSER', DEFAULT_BROWSER)
        ).lower()

        def get_browser(func, *args, **kwargs):
            try:
                return func(*args, **kwargs)
            except POLLING_EXCEPTIONS:
                return None

        def get_local_browser(browser_name):
            method = getattr(self, browser_name, None)

            if method:
                return method()

            raise SeleniumExError(
                'Incorrect browser name "{}"'.format(browser_name),
            )

        delay = delay or self.__config.get('POLLING_DELAY')
        timeout = timeout or self.__config.get('POLLING_TIMEOUT') or GET_DRIVER_TIMEOUT

        if self.__config.get('USE_REMOTE', False):
            driver = waiting_for(
                lambda: get_browser(self.remote, self.__browser_name),
                delay=delay,
                timeout=timeout,
                exc_cls=SeleniumExError,
                message='Browser "{}" have not been started for "{}" sec.'.format(
                    self.__browser_name, timeout,
                ),
            )
        else:
            driver = waiting_for(
                lambda: get_browser(get_local_browser, self.__browser_name),
                delay=delay,
                timeout=timeout,
                exc_cls=SeleniumExError,
                message='Browser "{}" have not been started for "{}" sec.'.format(
                    self.__browser_name, timeout,
                ),
            )

        return driver


class SeleniumAssertion(case.AssertionBase):

    DEFAULT_TIMEOUT = 0.5

    def any_text_in_page(self, browser, texts, msg=None):
        page_text = browser.text
        error_message = u'Not of those texts was not found on page "{}". Texts: [{}]'

        for text in texts:
            if text in page_text:
                break
        else:
            error_message = error_message.format(
                browser.driver.current_url,
                u', '.join(texts),
            )
            self.fail(msg or error_message)

    def text_in_page(self, browser, text_or_texts, timeout=None, msg=None):
        error_message = u'Text "{}" not found on page "{}"'

        def check_text(txt):
            try:
                return txt in browser.text
            except (HTTPException, StaleElementReferenceException):
                return False

        if isinstance(text_or_texts, (list, tuple)):
            for t in text_or_texts:
                waiting_for(
                    check_text,
                    args=(t,),
                    exc_cls=AssertionError,
                    message=msg or error_message.format(
                        t, browser.driver.current_url,
                    ),
                    delay=browser.config.POLLING_DELAY,
                    timeout=timeout or browser.config.POLLING_TIMEOUT or self.DEFAULT_TIMEOUT,
                )
        else:
            waiting_for(
                check_text,
                args=(text_or_texts,),
                exc_cls=AssertionError,
                message=msg or error_message.format(
                    text_or_texts, browser.driver.current_url,
                ),
                delay=browser.config.POLLING_DELAY,
                timeout=timeout or browser.config.POLLING_TIMEOUT or self.DEFAULT_TIMEOUT,
            )

    def web_element_exist(self, browser, query, timeout=None, msg=None):
        waiting_for(
            lambda: browser.query.form_object(query).exist,
            exc_cls=AssertionError,
            message=msg or u'Web element was not found on page "{}"'.format(
                browser.driver.current_url,
            ),
            delay=browser.config.POLLING_DELAY,
            timeout=timeout or browser.config.POLLING_TIMEOUT or self.DEFAULT_TIMEOUT,
        )

    def web_element_not_exist(self, browser, query, timeout=None, msg=None):
        waiting_for(
            lambda: not browser.query.form_object(query).exist,
            exc_cls=AssertionError,
            message=msg or u'Web element was found on page "{}"'.format(
                browser.driver.current_url,
            ),
            delay=browser.config.POLLING_DELAY,
            timeout=timeout or browser.config.POLLING_TIMEOUT or self.DEFAULT_TIMEOUT,
        )

    def content_in_web_element_exist(self, browser, query_object, timeout=None, msg=None):
        waiting_for(
            lambda: len(browser.query.form_object(query_object).first().text) > 0,
            exc_cls=AssertionError,
            message=msg or 'Content does not exist inside element on page "{}"'.format(
                browser.driver.current_url,
            ),
            delay=browser.config.POLLING_DELAY,
            timeout=timeout or browser.config.POLLING_TIMEOUT or self.DEFAULT_TIMEOUT,
        )


assertion = SeleniumAssertion()


def get_selenium_from_case(self):
    return self._SeleniumCase__selenium


def case_of_browsers(*browsers):
    def wrapper(cls):
        setattr(cls, '__browsers__', browsers)
        return cls
    return wrapper


def inject_driver(f):
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        return f(
            self, get_selenium_from_case(self).browser, *args, **kwargs
        )
    return wrapper


class SeleniumCaseLayer(case.CaseLayer):

    def on_require(self, require):
        if EX_NAME not in require:
            require.append(EX_NAME)

    def on_setup(self, case):
        if not case.__browsers__ and not case.config.SELENIUM_BROWSERS:
            get_selenium_from_case(case).start()

    def on_teardown(self, case):
        selenium = get_selenium_from_case(case)

        if selenium.browser.config.LOGS_PATH:
            save_logs(case, selenium)

        selenium.stop()


class SeleniumCase(case.Case):

    __browsers__ = None
    __layers__ = (SeleniumCaseLayer(), )
    __assertion_class__ = SeleniumAssertion

    def __init__(self, *args, **kwargs):
        # do it latter
        kwargs.update(use_flows=False)

        super(SeleniumCase, self).__init__(*args, **kwargs)

        self.__selenium = self.ext('selenium')

        if steps.is_step_by_step_case(self):
            step_methods = []

            for step_method in steps.get_step_methods(self):
                step_methods.append(inject_driver(step_method))

            setattr(self.__class__, steps.STEPS_STORAGE_ATTRIBUTE_NAME, step_methods)
        else:
            method = getattr(self.__class__, runnable.method_name(self))
            setattr(self.__class__, runnable.method_name(self), inject_driver(method))

            case.apply_flows(self)

    def __reason__(self):
        try:
            if self.__selenium.browser:  # if browser wasn't init, that reason can't be created
                screen_url = self.__selenium.config.get('SCREEN_URL', None)
                screen_path = self.__selenium.config.get('SCREEN_PATH', None)

                if screen_path:
                    try:
                        file_name = random_file_name('.png')
                        screen_path = os.path.join(screen_path, file_name)

                        with self.__selenium.browser.disable_polling():
                            if self.__selenium.browser.get_screenshot_as_file(screen_path):
                                if screen_url:
                                    self.__selenium.browser.reason_storage['screen url'] = u'{}{}'.format(
                                        screen_url, file_name,
                                    )
                                else:
                                    self.__selenium.browser.reason_storage['screen path'] = screen_path
                    except BaseException as error:
                        logger.warn(error, exc_info=True)

                return super(SeleniumCase, self).__reason__() + reason.create_item(
                    'Selenium',
                    'info from selenium extension',
                    *(u'{}: {}'.format(k, v) for k, v in self.__selenium.browser.reason_storage.items())
                )

            return super(SeleniumCase, self).__reason__()
        except BaseException as error:
            logger.warn(error, exc_info=True)
            return super(SeleniumCase, self).__reason__()

    def __repeat__(self):
        if self.config.SELENIUM_BROWSERS:
            for browser_name in self.config.SELENIUM_BROWSERS:
                if self.__browsers__ and browser_name not in self.__browsers__:
                    continue
                if self.__selenium.browser and self.__selenium.browser.config.LOGS_PATH:
                    save_logs(self, self.__selenium)

                self.__selenium.stop()
                self.__selenium.start(browser_name)
                yield
        elif self.__browsers__:
            for browser_name in self.__browsers__:
                if self.__selenium.browser and self.__selenium.browser.config.LOGS_PATH:
                    save_logs(self, self.__selenium)

                self.__selenium.stop()
                self.__selenium.start(browser_name)
                yield
        else:
            yield


class SeleniumSuiteLayer(suite.SuiteLayer):

    def on_require(self, require):
        if EX_NAME not in require:
            require.append(EX_NAME)


class SeleniumSuite(suite.Suite):

    __case_class__ = SeleniumCase
    __layers__ = (SeleniumSuiteLayer(), )



# -*- coding: utf8 -*-

import time
import socket
import logging
from functools import wraps

try:
    from httplib import HTTPException
except ImportError:  # please python 3
    from http.client import HTTPException

from selenium.common.exceptions import WebDriverException

from ...utils import pyv
from .exceptions import SeleniumExError


logger = logging.getLogger(__name__)


POLLING_EXCEPTIONS = (
    IOError,
    OSError,
    HTTPException,
    WebDriverException,
)
DEFAULT_POLLING_TIMEOUT = 30


def do(callback=None, timeout=DEFAULT_POLLING_TIMEOUT, delay=None):
    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            exc = None
            t_start = time.time()

            while time.time() <= t_start + timeout:
                try:
                    logger.debug(
                        u'Do polling, try to call "{}", args={}, kwargs={}'.format(
                            pyv.get_func_name(f), pyv.unicode(args), pyv.unicode(kwargs),
                        ),
                    )
                    return f(*args, **kwargs)
                except socket.error as error:  # if connection will be refused,
                    # that we not need to continue polling
                    raise error
                except POLLING_EXCEPTIONS as error:
                    exc = error
                    if delay:
                        time.sleep(delay)
                    continue
            else:
                if exc:
                    raise exc
                raise SeleniumExError(
                    'Timeout {} exceeded'.format(timeout),
                )

        return wrapped

    if callback is not None:
        return wrapper(callback)

    return wrapper



# -*- coding: utf-8 -*-

from functools import wraps
from types import MethodType
from collections import OrderedDict
from contextlib import contextmanager

from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.common.action_chains import ActionChains

from . import polling
from .router import Router
from .query import QueryProcessor
from ...utils.common import waiting_for
from .utils import change_name_from_python_to_html


METHOD_ALIASES = {
    'go_to': 'get',
    'set': 'send_keys',
}


def _check_equal(we, attr, value):
    if attr == 'text':
        return we.text == value

    return getattr(we.attr, attr) == value


def factory_method(f, driver, config, allow_polling=True, reason_storage=None):
    @wraps(f)
    def wrapper(*args, **kwargs):
        result = f(*args, **kwargs)

        if isinstance(result, WebElement):
            return WebElementProxy(
                result,
                config=config,
                driver=driver,
                allow_polling=allow_polling,
                reason_storage=reason_storage,
            )

        if isinstance(result, (list, tuple)):
            we_list = WebElementListProxy(
                driver,
                config,
                reason_storage=reason_storage,
            )

            for obj in result:
                if isinstance(obj, WebElement):
                    we_list.append(
                        WebElementProxy(
                            obj,
                            config=config,
                            allow_polling=allow_polling,
                            reason_storage=reason_storage,
                        ),
                    )
                else:
                    return result

            return we_list
        return result
    return wrapper


class WebElementListProxy(list):

    def __init__(self, driver, config, reason_storage=None):
        super(WebElementListProxy, self).__init__()

        self.__driver = driver
        self.__config = config
        self.__reason_storage = reason_storage or OrderedDict()

    @property
    def driver(self):
        return self.__driver

    @property
    def config(self):
        return self.__config

    @property
    def query(self):
        return self.__driver.query

    @property
    def reason_storage(self):
        return self.__reason_storage

    def get_by(self, **kwargs):
        try:
            return next(
                we for we in self
                if all(_check_equal(we, k, v) for k, v in kwargs.items())
            )
        except StopIteration:
            pass

    def filter(self, **kwargs):
        for we in self:
            if all(_check_equal(we, k, v) for k, v in kwargs.items()):
                yield we


class BaseProxy(object):

    def __init__(self, wrapped, config=None, driver=None, reason_storage=None, allow_polling=True):
        self.__driver = driver
        self.__config = config
        self.__wrapped = wrapped
        self.__query = QueryProcessor(self)
        self.__allow_polling = allow_polling
        self.__reason_storage = reason_storage or OrderedDict()

        assert self.__class__ != BaseProxy, 'This is base class only. Can not be instanced.'

    def __repr__(self):
        return repr(self._wrapped)

    def __getattr__(self, item):
        attr = getattr(
            self._wrapped, METHOD_ALIASES.get(item, item), None,
        )

        if attr is None:
            return getattr(self.query, item)

        if callable(attr) and type(attr) == MethodType:
            if self.allow_polling:
                return polling.do(
                    callback=factory_method(
                        attr,
                        self.driver,
                        self.config,
                        allow_polling=self.allow_polling,
                        reason_storage=self.reason_storage,
                    ),
                    delay=self.config.POLLING_DELAY,
                    timeout=self.config.POLLING_TIMEOUT,
                )

            return factory_method(
                attr,
                self.driver,
                self.config,
                allow_polling=self.allow_polling,
                reason_storage=self.reason_storage,
            )

        return attr

    @property
    def _wrapped(self):
        return self.__wrapped

    @property
    def config(self):
        return self.__config

    @property
    def query(self):
        return self.__query

    @property
    def driver(self):
        return self.__driver

    @property
    def allow_polling(self):
        return (
            bool(self.config.POLLING_TIMEOUT) and self.__allow_polling
        )

    @property
    def reason_storage(self):
        return self.__reason_storage

    @property
    def text(self):
        def get_text():
            if isinstance(self._wrapped, WebElement):
                return self._wrapped.text
            return self._wrapped.find_element_by_tag_name('body').text

        if self.allow_polling:
            func = polling.do(
                get_text,
                delay=self.config.POLLING_DELAY,
                timeout=self.config.POLLING_TIMEOUT,
            )
            return func()

        return get_text()

    @contextmanager
    def polling(self, func=None, exc_cls=None, message=None, timeout=None, args=None, kwargs=None):
        to_restore = self.__allow_polling
        self.__allow_polling = True

        if func:
            waiting_for(
                func,
                args=args,
                kwargs=kwargs,
                exc_cls=exc_cls,
                message=message,
                delay=self.config.POLLING_DELAY,
                timeout=timeout or self.config.POLLING_TIMEOUT or 0.5,
            )

        try:
            yield
        finally:
            self.__allow_polling = to_restore

    @contextmanager
    def disable_polling(self):
        try:
            self.__allow_polling = False
            yield
        finally:
            self.__allow_polling = True


class WebElementProxy(BaseProxy):

    def __init__(self, *args, **kwargs):
        super(WebElementProxy, self).__init__(*args, **kwargs)

        assert isinstance(self._wrapped, WebElement), 'This is proxy to WebElement only'

    @property
    def css(self):
        return WebElementCssToObject(self)

    @property
    def attr(self):
        return WebElementToObject(self, allow_raise=False)


class WebDriverProxy(BaseProxy):

    def __init__(self, *args, **kwargs):
        super(WebDriverProxy, self).__init__(*args, **kwargs)

        assert isinstance(self._wrapped, WebDriver), 'This is proxy to WebDriver only'

        self.__router = Router(self)
        self.__action_chains = ActionChainsProxy(self)

    @property
    def driver(self):
        return self

    @property
    def router(self):
        return self.__router

    @property
    def action_chains(self):
        return self.__action_chains

    @property
    def current_url(self):
        if self.allow_polling:
            func = polling.do(
                lambda: self._wrapped.current_url,
                delay=self.config.POLLING_DELAY,
                timeout=self.config.POLLING_TIMEOUT,
            )
            return func()

        return self._wrapped.current_url


class ActionChainsProxy(object):

    def __init__(self, proxy):
        self.__action_chains = ActionChains(proxy.driver)

    def __call__(self, proxy):
        return self.__class__(proxy.driver)

    def __getattr__(self, item):
        return getattr(self.__action_chains, item)

    def __repr__(self):
        return repr(self.__action_chains)

    def reset(self):
        self.__action_chains._actions = []

    def perform(self, reset=True):
        self.__action_chains.perform()

        if reset:
            self.reset()


class WebElementToObject(object):

    def __init__(self, proxy, allow_raise=True):
        self.__dict__['__proxy__'] = proxy
        self.__dict__['__allow_raise__'] = allow_raise

    def __getattr__(self, item):
        atr = self.__dict__['__proxy__'].get_attribute(
            change_name_from_python_to_html(item),
        )

        if atr:
            return atr

        if self.__dict__['__allow_raise__']:
            raise AttributeError(item)

    def __setattr__(self, key, value):
        self.__dict__['__proxy__'].parent.execute_script(
            'arguments[0].setAttribute(arguments[1], arguments[2]);',
            self.__dict__['__proxy__'],
            change_name_from_python_to_html(key),
            value
        )


class WebElementCssToObject(object):

    def __init__(self, proxy):
        self.__dict__['__proxy__'] = proxy

    def __getattr__(self, item):
        return self.__dict__['__proxy__'].value_of_css_property(
            change_name_from_python_to_html(item),
        )

    def __setattr__(self, key, value):
        self.__dict__['__proxy__'].parent.execute_script(
            'arguments[0].style[arguments[1]] = arguments[2];',
            self.__dict__['__proxy__'],
            change_name_from_python_to_html(key),
            value
        )



# -*- coding: utf-8 -*-

import time

from ..exceptions import TimeoutException


WAITING_FOR_TIMEOUT = 30


def waiting_for(func, timeout=None, exc_cls=None, message=None, delay=None, args=None, kwargs=None):
    args = args or tuple()
    kwargs = kwargs or {}

    timeout = timeout or WAITING_FOR_TIMEOUT

    t_start = time.time()

    while time.time() <= t_start + timeout:
        result = func(*args, **kwargs)

        if result:
            return result

        if delay:
            time.sleep(delay)
    else:
        message = message or 'Timeout {} exceeded'.format(timeout)
        if exc_cls:
            raise exc_cls(message)
        raise TimeoutException(message)


def call_to_chain(chain, method_name, *args, **kwargs):
    for obj in chain:
        if method_name:
            getattr(obj, method_name)(*args, **kwargs)
        else:
            obj(*args, **kwargs)


def measure_time():
    start_time = time.time()
    return lambda: time.time() - start_time



# -*- coding: utf-8 -*-

from six import with_metaclass

from ...utils import pyv
from .query import QueryObject


def page_object_factory(page_element):
    def wrapper(self):
        if page_element.obj_class:
            result = page_element.get_instance(self.area)
        else:
            we = None

            for query_object in page_element.query_set[:-1]:
                if we:
                    we = we.query.from_object(
                        query_object,
                    ).first()
                else:
                    we = self.query.from_object(
                        query_object,
                    ).first()

            if we:
                result = we.query.from_object(page_element.query_set[-1:][0])
            else:
                result = self.query.from_object(page_element.query_set[-1:][0])

            if page_element.wait_timeout:
                result.wait(
                    page_element.wait_timeout,
                )

            if page_element.is_list:
                result = result.all()
            elif page_element.index:
                result = result.get(
                    page_element.index,
                )
            else:
                result = result.first()

        if page_element.proxy:
            if pyv.is_class_type(page_element.proxy):
                return page_element.proxy(result)
            return lambda: page_element.proxy(result)
        return result

    return property(wrapper)


class PageObjectProxy(object):

    def __init__(self, proxy):
        self.__wrapped = proxy

    def __getattr__(self, item):
        return getattr(self.__wrapped, item)

    def __repr__(self):
        return repr(self.__wrapped)

    @property
    def _wrapped(self):
        return self.__wrapped

    @property
    def driver(self):
        return self.__wrapped.driver


class PageObject(object):

    def __init__(self, *args, **options):
        assert bool(args), 'Element can not be created without args'

        self.__query_set = None
        self.__obj_class = None
        self.__obj_instance = None

        if pyv.is_class_type(args[0]):
            self.__obj_class = args[0]
        if isinstance(args[0], PageObject):
            self.__query_set = args[0].query_set + args[1:]
        else:
            self.__query_set = args

        self.__index = options.get('index', None)
        self.__proxy = options.get('proxy', None)
        self.__is_list = options.get('is_list', False)
        self.__wait_timeout = options.get('wait_timeout', None)

    def __call__(self, *args, **kwargs):
        # for IDE only
        raise TypeError(
            '"{}" object is not callable'.format(self.__class__.__name__),
        )

    def __getattr__(self, item):
        # for IDE only
        raise AttributeError(item)

    @property
    def obj_class(self):
        return self.__obj_class

    def get_instance(self, area):
        if self.__obj_instance is None:
            self.__obj_instance = self.__obj_class(area)
        return self.__obj_instance

    @property
    def query_set(self):
        return self.__query_set

    @property
    def index(self):
        return self.__index

    @property
    def is_list(self):
        return self.__is_list

    @property
    def proxy(self):
        return self.__proxy

    @proxy.setter
    def proxy(self, cls):
        assert issubclass(cls, PageObjectProxy), \
            '"{}" is not "PageObjectProxy" subclass'.format(
                cls.__name__,
        )
        self.__proxy = cls

    @property
    def wait_timeout(self):
        return self.__wait_timeout


class PageApi(object):

    def __init__(self, page):
        self.__page = page

    @property
    def page(self):
        return self.__page


class PageMeta(type):
    """
    Factory for creating page object class
    """

    def __new__(mcs, name, bases, dct):
        cls = type.__new__(mcs, name, bases, dct)

        cls_items = (
            (a, getattr(cls, a, None))
            for a in dir(cls)
            if not a.startswith('_')
        )

        for atr, value in cls_items:
            if isinstance(value, PageObject):
                setattr(cls, atr, page_object_factory(value))

        return cls


class Page(with_metaclass(PageMeta, object)):

    __wrapper__ = None
    __api_class__ = PageApi

    def __init__(self, proxy):
        self.__proxy = proxy
        self.__api = self.__api_class__(self)

    def __getattr__(self, item):
        return getattr(self.area, item)

    @property
    def api(self):
        return self.__api

    @property
    def area(self):
        if self.__wrapper__:
            if not isinstance(self.__wrapper__, QueryObject):
                raise TypeError(
                    '"__wrapper__" can be instance of QueryObject only',
                )
            return self.__proxy.query.from_object(
                self.__wrapper__,
            ).first()

        return self.__proxy

    @property
    def driver(self):
        return self.__proxy.driver

    @property
    def query(self):
        return self.area.query

    def relate_to(self, proxy):
        self.__proxy = proxy

    def refresh(self):
        self.__proxy.driver.refresh()
